import { createHmac, createHash, randomBytes } from 'node:crypto'
import dayjs from 'dayjs'

import SessionStore from './sessionStore.ts'

/**
 *  Session
 * 
 *  @author Björn Hase <me@herr-hase.wtf>
 *  @license http://opensource.org/licenses/MIT The MIT License
 *  @link https://git.node001.net/nano/session.git
 *
 */
class Session {

    // database object 
    private db

    // store for sessions
    private sessionStore

    // max logins for one user 
    private maxLogins = 5;

    // duration for expired_at
    private duration = 60;

    /**
     *
     *  @param object db
     *
     */
    constructor(db: object) {
        this.db = db
        this.sessionStore = new SessionStore(db)

        if (!process.env.NANO_SESSION_SECRET) {
            throw new Error('NANO_SESSION_SECRET is undefined, please check .env-file')
        }

        // if NANO_MAX_LOGINS is set in .env
        if (process.env.NANO_SESSION_MAX_LOGINS) {
            this.maxLogins = process.env.NANO_SESSION_MAX_LOGINS
        }

        // if NANO_SESSION_DURATION is set in .env
        if (process.env.NANO_SESSION_DURATION) {
            this.duration = process.env.NANO_SESSION_DURATION
        }
    }

    /**
     *  create session 
     *
     *  @param number userId
     *  @param string agent
     *  @param string language
     *  @param string ip
     *  @return object
     *
     */
    public async create(userId: number, agent: string, language: string, ip = ''): object {

        const userHash = this.createUserHash(userId)
        const userHmac = this.createUserHmac(userHash)
        const users = await this.sessionStore.findOneByUser(userHmac)

        // if users are not longer can login, drop oldest
        if (users && users.length > this.maxLogins) {
            const session = await this.sessionStore.dropOldest(userHash)
        }

        const token = randomBytes(64).toString('hex')
        const tokenHmac = this.createTokenHmac(token, agent, language, ip)

        // create date 
        const date = dayjs()

        await this.sessionStore.create({
            'user_id': userId,
            'user' : userHmac,
            'token' : tokenHmac,
            'expired_at': date.add(this.duration, 'minute').toISOString(),
            'created_at': date.toISOString()
        })

        return {
            'user': userHash,
            'token': token    // raw token goes in the cookie 
        }
    }

    /**
     *  delete single session for a user
     *
     *  @param object userId
     *
     */
    public async destroy(sessionId: number): void {
        await this.sessionStore.remove(sessionId)
    }

    /**
     *  delete all session for a user
     *
     *  @param number userId
     *
     */
    public async destroyAll(userId: number): void {
        const userHmac = this.createUserHmac(this.createUserHash(userId))
        const sessions = await this.sessionStore.findByUser(userHmac)

        for (const index in sessions) {
            await this.destroy(sessions[index].id)
        }
    }

    /**
     *  getting single session
     *
     *  @param number userId
     *  @param string agent
     *  @param string language
     *  @param string ip
     *  @param string token
     *  @return string 
     *
     */
    public async get(userHash: string, token: string, agent: string, language: string, ip = ''): string {
        const userHmac = this.createUserHmac(userHash)
        const tokenHmac = this.createTokenHmac(token, agent, language, ip)

        let session = await this.sessionStore.findOneByUserAndToken(userHmac, tokenHmac)

        // if user found, check expired_at, if expired, delete session
        if (session && session.expired_at) {
            if (dayjs().isAfter(dayjs(session.expired_at))) {
                await this.destroy(session.id)
                session = null
            }
        }

        return session
    }

    /**
     *  creating hmac for userId 
     *
     *  @param number userId
     *  @return string 
     *
     */
    private createUserHash(userId: number): string {
        const hash = createHash('sha256')
        const buffer = new Buffer(userId)
        
        return hash.update(buffer).digest('hex')
    }

    /**
     *  creating hmac for userId 
     *
     *  @param number userId
     *  @return string 
     *
     */
    private createUserHmac(userHash: string): string {
        const hmac = createHmac('sha512', process.env.NANO_SESSION_SECRET)
        const buffer = new Buffer(userHash)
        
        return hmac.update(buffer).digest('base64')
    }

    /**
     *  creating hmac for token 
     *
     *  @param number userId
     *  @param string agent
     *  @param string language
     *  @param string ip
     *  @return string 
     *
     */
    private createTokenHmac(token: string, agent: string, language: string, ip = ''): string {
        const hmac = createHmac('sha512', process.env.NANO_SESSION_SECRET)
        const buffer = new Buffer(token + agent + language + ip)

        return hmac.update(buffer).digest('base64') 
    }
}

export default Session
